#pragma comment(linker, "/version:1.1") // MUST be in the form of major.minor

#include <windows.h>
#include <shlwapi.h>
#include <commctrl.h>
#include "libs\Wow64.h"

#define SZ_TITLE TEXT("6.1.7601 USB Fixup Utility")

#define countof(x) (sizeof(x)/sizeof(x[0]))

/* Set a "feather": if the current file in drivers is newer than or within 48
 * hours of the timestamp of the file in WinSxS, consider the file up-to-date.
 */
#define TIME_FEATHER 1728000000000 // 48 * 3600 * 10,000,000 = 1,728,000,000,000

/* Interchanging FILETIME with UINT64 is bad because it might lead to the UINT64
 * being misaligned, which is bad for Itanium and other uarchs (but not for x86
 * or x64) (though, in the usage in this particular program, such a misalignment
 * is not possible) and because it assumes that the uarch is little-endian (but
 * since every uarch that 7601 runs on is LE, this, too, is moot).
 */
typedef union {
	FILETIME ft;
	UINT64 ui;
} FILETIMEEX, *PFILETIMEEX;

UINT64 __fastcall GetTimeByName( PCTSTR pszPath )
{
	HANDLE hFile = CreateFile(
		pszPath,
		0,
		FILE_SHARE_READ | FILE_SHARE_WRITE,
		NULL,
		OPEN_EXISTING,
		FILE_FLAG_BACKUP_SEMANTICS,
		NULL
	);

	FILETIMEEX ftex;
	ftex.ui = 0;

	if (hFile != INVALID_HANDLE_VALUE)
	{
		GetFileTime(hFile, NULL, NULL, &ftex.ft);
		CloseHandle(hFile);
	}

	return(ftex.ui);
}

__forceinline
BOOL CheckWindowsVersion( )
{
	OSVERSIONINFO osvi;
	osvi.dwOSVersionInfoSize = sizeof(osvi);

	return(
		GetVersionEx(&osvi) &&
		osvi.dwMajorVersion == 6 &&
		osvi.dwMinorVersion == 1 &&
		osvi.dwBuildNumber == 7601
	);
}

__forceinline
VOID EnableProtectedPendmoves( )
{
	// Windows automatically deletes this registry value after the system
	// processes pending file operations after reboot.

	static const TCHAR szSM[] = TEXT("SYSTEM\\CurrentControlSet\\Control\\Session Manager");
	static const TCHAR szAPR[] = TEXT("AllowProtectedRenames");

	HKEY hKey = NULL;

	if ( RegOpenKeyEx(HKEY_LOCAL_MACHINE, szSM, 0, KEY_SET_VALUE, &hKey) == ERROR_SUCCESS &&
		 hKey != NULL )
	{
		DWORD dwData = 1;
		RegSetValueEx(hKey, szAPR, 0, REG_DWORD, (PBYTE)&dwData, sizeof(DWORD));
		RegCloseKey(hKey);
	}
}

__forceinline
VOID RebootWindows( )
{
	HANDLE hToken;

	if (OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken))
	{
		TOKEN_PRIVILEGES tkp;

		LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, &tkp.Privileges[0].Luid);
		tkp.PrivilegeCount = 1;
		tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
		AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, (PTOKEN_PRIVILEGES) NULL, 0);

		ExitWindowsEx(EWX_REBOOT, 0);

		tkp.Privileges[0].Attributes = 0;
		AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, (PTOKEN_PRIVILEGES) NULL, 0);

		CloseHandle(hToken);
	}
}

static const PCTSTR pszFiles[][3] =
{
	{
		TEXT("usbehci.sys"),
		TEXT("x86_usbport.inf_31bf3856ad364e35_6.1.7601.17514_none_bfc9c95e61cfba61"),
		TEXT("amd64_usbport.inf_31bf3856ad364e35_6.1.7601.17514_none_1be864e21a2d2b97"),
	},

	{
		TEXT("usbport.sys"),
		TEXT("x86_usbport.inf_31bf3856ad364e35_6.1.7601.17514_none_bfc9c95e61cfba61"),
		TEXT("amd64_usbport.inf_31bf3856ad364e35_6.1.7601.17514_none_1be864e21a2d2b97"),
	},

	{
		TEXT("winusb.sys"),
		TEXT("x86_winusb.inf_31bf3856ad364e35_6.1.7601.17514_none_f9fc4e7173e3735c"),
		TEXT("amd64_winusb.inf_31bf3856ad364e35_6.1.7601.17514_none_561ae9f52c40e492"),
	},
};

#pragma comment(linker, "/entry:main")
void main( )
{
	BOOL fWin64, fNeedReboot = FALSE;

	TCHAR szTemp[MAX_PATH];
	TCHAR szMessage[0x200];

	InitCommonControls();

	if (!CheckWindowsVersion())
	{
		MessageBox(NULL, TEXT("You are not running 6.1.7601."), NULL, MB_ICONERROR);
		ExitProcess(~0);
	}

	EnableProtectedPendmoves();

	if (fWin64 = Wow64CheckProcess())
		Wow64DisableFsRedir();

	if ( GetSystemWindowsDirectory(szTemp, countof(szTemp)) &&
	     SetCurrentDirectory(szTemp) )
	{
		UINT i, idxSource = (fWin64) ? 2 : 1;

		TCHAR szSource[MAX_PATH];
		TCHAR szTarget[MAX_PATH];

		DWORD cchSourceRoot = StrCatChainW(szSource, countof(szSource), 0, TEXT("WinSxS\\"));
		DWORD cchTargetRoot = StrCatChainW(szTarget, countof(szTarget), 0, TEXT("System32\\drivers\\"));

		szMessage[0] = 0;

		for (i = 0; i < countof(pszFiles); ++i)
		{
			UINT64 uiTimeSource, uiTimeTarget;
			PCTSTR pszStatus;

			StrCatChainW(szSource, countof(szSource), cchSourceRoot, pszFiles[i][idxSource]);
			StrCatChainW(szSource, countof(szSource), -1, TEXT("\\"));
			StrCatChainW(szSource, countof(szSource), -1, pszFiles[i][0]);

			StrCatChainW(szTarget, countof(szTarget), cchTargetRoot, pszFiles[i][0]);

			StrCatChainW(szTemp, countof(szTemp), 0, szTarget);
			StrCatChainW(szTemp, countof(szTemp), -1, TEXT(".usbfix"));

			uiTimeSource = GetTimeByName(szSource);
			uiTimeTarget = GetTimeByName(szTarget);

			if (uiTimeTarget == 0)
				pszStatus = TEXT("File is not used by the system.");
			else if (uiTimeSource == 0)
				pszStatus = TEXT("Cannot locate the source file in WinSxS.");
			else if (uiTimeTarget + TIME_FEATHER >= uiTimeSource)
				pszStatus = TEXT("File is already up-to-date.");
			else if (!CopyFile(szSource, szTemp, FALSE))
				pszStatus = TEXT("Error encountered while creating temporary staging file.");
			else if (!MoveFileEx(szTemp, szTarget, MOVEFILE_REPLACE_EXISTING | MOVEFILE_DELAY_UNTIL_REBOOT))
				pszStatus = TEXT("Error encountered while scheduling reboot file operation.");
			else
			{
				pszStatus = TEXT("File replacement will be carried out upon reboot.");
				fNeedReboot = TRUE;
			}

			StrCatChainW(szMessage, sizeof(szMessage), -1, pszFiles[i][0]);
			StrCatChainW(szMessage, sizeof(szMessage), -1, TEXT(":\t"));
			StrCatChainW(szMessage, sizeof(szMessage), -1, pszStatus);
			StrCatChainW(szMessage, sizeof(szMessage), -1, TEXT("\n"));
		}

		if (fNeedReboot)
			StrCatChainW(szMessage, sizeof(szMessage), -1, TEXT("\nSave your work and press OK to reboot your computer."));
		else
			StrCatChainW(szMessage, sizeof(szMessage), -1, TEXT("\nNo changes were made to your system."));

		MessageBox(NULL, szMessage, SZ_TITLE, MB_ICONINFORMATION);

		if (fNeedReboot)
			RebootWindows();

		ExitProcess(1);
	}
	else
	{
		MessageBox(NULL, TEXT("Cannot find the Windows directory."), NULL, MB_ICONERROR);
		ExitProcess(~0);
	}
}
